(* Driver program *)

open Printf

let buffer = String.create 16384

let rec copy_file ic oc =
  let n = input ic buffer 0 16384 in
  if n = 0 then () else begin
    output oc buffer 0 n;
    copy_file ic oc
  end

let safe_remove f =
  try Sys.remove f with _ -> ()

let filesize f =
  (Unix.stat f).Unix.st_size

exception Time_is_up

let _ =
  (* Arm timer for 10 seconds before the deadline *)
  let timeout =
    if Array.length Sys.argv >= 2 then int_of_string Sys.argv.(1) else 180 in  
  Sys.set_signal Sys.sigalrm (Sys.Signal_handle (fun _ -> raise Time_is_up));
  if timeout > 10 then ignore (Unix.alarm (timeout - 10));
  (* Copy input to temp file *)
  let input_name = Filename.temp_file "temp" ".smlng" in
  let oc = open_out input_name in
  copy_file stdin oc;
  close_out oc;
  let best_result_so_far = ref input_name in
  let files_to_clean = ref [input_name] in
  fprintf stderr "Input file: size %d\n" (filesize input_name);
  flush stderr;
  try
    (* Optimize input *)
    let input_data = Parser.parse_file input_name in
    for round = 1 to 5 do
      let output_name = Filename.temp_file "temp" ".smlng" in
      let oc = open_out output_name in
      files_to_clean := output_name :: !files_to_clean;
      Opti.transcribe oc input_data;
      close_out oc;
      let output_size = filesize output_name in
      fprintf stderr "Round %d: result of size %d" round output_size;
      if output_size >= filesize !best_result_so_far then begin
        prerr_endline ", nothing better, discarded"
      end else if not (Validate.equivalent input_name output_name) then begin
        prerr_endline ", NOT EQUIVALENT, discarded"
      end else begin
        best_result_so_far := output_name;
        prerr_endline ", kept"
      end;
      let output_name = Filename.temp_file "temp" ".smlng" in
      let oc = open_out output_name in
      files_to_clean := output_name :: !files_to_clean;
      Opti.rev_transcribe oc input_data;
      close_out oc;
      let output_size = filesize output_name in
      fprintf stderr "Round %d reversed: result of size %d" round output_size;
      if output_size >= filesize !best_result_so_far then begin
        prerr_endline ", nothing better, discarded"
      end else if not (Validate.equivalent input_name output_name) then begin
        prerr_endline ", NOT EQUIVALENT, discarded"
      end else begin
        best_result_so_far := output_name;
        prerr_endline ", kept"
      end;
      (* We have so time to spare, so try bumping up the parameters of the
         optimizer.  Who knowns, a better result might be found...
         Moreover, this will annoy the judges to no end!!! *)
      Opti.num_solutions := 2 * !Opti.num_solutions;
      Opti.num_states := 2 * !Opti.num_states
    done;
    (* We don't know what to do now, so let's output the result *)
    ignore (Unix.alarm 0);
    raise Time_is_up
  with x ->
    (* Whatever happened before, it's time to output the result! *)
    let ic = open_in !best_result_so_far in
    copy_file ic stdout;
    close_in ic;
    List.iter safe_remove !files_to_clean;
    match x with 
      Time_is_up -> exit 0
    | Lexer.Lexical_error msg -> 
        fprintf stderr "Lexical error: %s\n" msg; exit 2
    | Parser.Parse_error msg -> 
        fprintf stderr "Syntax error: %s\n" msg; exit 2
    | _ -> raise x
