open Lexer
open Parser

(* A state is a context + a stack of opened tags and contexts before each
   tag was opened *)

type state =
  { ctx: context;
    stack: (tag * context) list }

(* Given a state and a target context, generate sequences of
   open/close tags that result into the target context. *)

type action = Open of tag | Close of tag

let tag_size = function
    EM -> 2
  | PL -> 2
  | TT -> 2
  | _ -> 1

let size_actions tags =
  let rec size accu = function
    [] -> accu
  | Open t :: rem -> size (accu + 2 + tag_size t) rem
  | Close t :: rem -> size (accu + 3 + tag_size t) rem
  in size 0 tags

let solutions = ref ([] : (int * action list * state) list)

let enter tags ctx stack =
  solutions :=
    (size_actions tags, List.rev tags, {ctx=ctx; stack=stack}) :: !solutions

let rec add_tags c d tags stack =
  if c = d then enter tags d stack else begin
    let try_tag t =
      add_tags (open_tag c t) d (Open t :: tags) ((t, c) :: stack) in
    if c.b = false && d.b = true then try_tag B;
    if c.em <> d.em then begin assert (not c.s); try_tag EM end;
    if c.i = false && d.i = true then try_tag I;
    if c.s = false && d.s = true then try_tag S;
    if c.tt = false && d.tt = true then try_tag TT;
    if c.u < d.u then try_tag U;
    if c.size <> d.size && d.size <> initial_size then
      try_tag (Size d.size);
    if c.color <> d.color && d.color <> initial_color then
      try_tag (Color d.color);
    if (c.s = true && d.s = false)
       || (c.em = true && d.em = false)
       || (c.i = true && d.i = false)
       || (c.b = true && d.b = false)
       || (c.tt = true && d.tt = false)
       || (c.u > d.u) then
      try_tag PL
  end

let rec pop_tags stk d tags =
  match stk with
    [] -> ()
  | (tag, c) :: rem ->
      add_tags c d (Close tag :: tags) rem;
      pop_tags rem d (Close tag :: tags)

(*
let rec apply_actions state = function
    [] -> state
  | Open t :: rem ->
      apply_actions {ctx = open_tag state.ctx t;
                     stack = (t, state.ctx) :: state.stack} rem
  | Close t :: rem ->
      match state.stack with
        (t', ctx) :: s when t = t' ->
          apply_actions {ctx = ctx; stack = s} rem
      | _ ->
          assert false
*)

let compare_solutions (cost1, _, _) (cost2, _, _) = cost1 - cost2

let num_solutions = ref 5

let transitions state dest =
  solutions := [];
  add_tags state.ctx dest [] state.stack;
  pop_tags state.stack dest [];
  (* Keep the 5 shortest solutions *)
  let sols =
    List2.truncate !num_solutions (List.sort compare_solutions !solutions) in
  solutions := [];
  sols

(* Explore various ways to render a list of runs. *)

let one_step_one_state run (cost, path, state) =
  List.map
    (fun (c, act, state') -> 
       (cost + c, (act, run.pos, run.len) :: path, state'))
    (transitions state run.Parser.ctx)

let rec closing_cost accu = function
    [] -> accu
  | (tag, _) :: rem -> closing_cost (accu + 3 + tag_size tag) rem

let size_path_state (cost, act, state) =
  cost + closing_cost 0 state.stack

let compare_path_state ps1 ps2 = size_path_state ps1 - size_path_state ps2

let num_states = ref 20

let one_step path_state_list run =
  List2.truncate !num_states
    (List.sort compare_path_state
      (List2.flat_map (one_step_one_state run) path_state_list))

let close_all_tags stk =
  List.map (fun (tag, _) -> Close tag) stk

let reconstitute runs =
  match
    List.fold_left one_step [0, [], {ctx=initial_context; stack=[]}] runs
  with
    [] -> assert false
  | (cost, path, final_state) :: _ ->
      List.rev ((close_all_tags final_state.stack, 0, 0) :: path)

(* Print the result *)

let print_result oc txt path =
  let print_tag = function
      B -> output_char oc 'B'
    | EM -> output_string oc "EM"
    | I -> output_char oc 'I'
    | PL -> output_string oc "PL"
    | S -> output_char oc 'S'
    | TT -> output_string oc "TT"
    | U -> output_char oc 'U'
    | Size s -> output_char oc s
    | Color c -> output_char oc c in
  let print_action = function
      Open t ->
        output_char oc '<'; print_tag t; output_char oc '>'
    | Close t ->
        output_string oc "</"; print_tag t; output_char oc '>' in
  List.iter
    (fun (act, start, len) ->
      List.iter print_action act; output oc txt start len)
    path

let transcribe oc (txt, runs) =
  print_result oc txt (reconstitute runs)


(* Same, but proceed right to left :-) *)

let invert_action = function Open t -> Close t | Close t -> Open t

let rec reverse_result accu laststart lastlen = function
    [] -> accu
  | (acts, start, len) :: rem ->
      reverse_result
        ((List.rev_map invert_action acts, laststart, lastlen) :: accu)
        start len rem

let rev_transcribe oc (txt, runs) =
  print_result oc txt (reverse_result [] 0 0 (reconstitute (List.rev runs)))
