(* Parse SML/NG.  Result is:
    - a string containing all characters to be output
    - a list of runs giving the context for the characters in the run *)

open Lexer

type context =
  { b: bool;
    em: bool;
    i: bool;
    s: bool;
    tt: bool;
    u: int;
    size: char;
    color: char }

type run =
  { pos: int;
    len: int;
    ctx: context }

let initial_color = '?'
let initial_size = '?'

let initial_context =
  { b = false; em = false; i = false; s = false;
    tt = false; u = 0; size = initial_size;
    color = initial_color }

(* Pseudo-output function that builds the runs *)

let text = Buffer.create 20
let runs = ref []
let last_was_space = ref false
let last_context = ref initial_context
let start_run = ref 0

let add_new_run () =
  let s = !start_run in
  let e = Buffer.length text in
  if s < e then begin
    let c = !last_context in
    let r = { pos = s; len = e - s; ctx = c } in
    runs := r :: !runs;
    start_run := e
  end
 
let same_space_context c d =
  c.tt = d.tt &&
  c.u = d.u &&
  c.size = d.size &&
  (c.u = 0 || c.color = d.color)

let output_char c ctx =
  (* If context changed, end the run *)
  if ctx <> !last_context then begin
    add_new_run ();
    last_context := ctx
  end;
  (* Add character *)
  Buffer.add_char text c;
  last_was_space := false

let output_blank c ctx =
  if same_space_context !last_context ctx then begin
    if ctx.tt then begin
      (* Add blank verbatim to current run *)
      Buffer.add_char text c;
      last_was_space := true
    end else if !last_was_space 
    then begin
      (* Just omit space (merged with preceding one) *)
      ()
    end else begin
      (* Add space to current run *)
      Buffer.add_char text ' ';
      last_was_space := true
    end
  end else begin
    (* Context changed, so end the run *)
    add_new_run ();
    last_context := ctx;
    if ctx.tt then begin
      Buffer.add_char text c;
      last_was_space := true
    end else begin
      Buffer.add_char text ' ';
      last_was_space := true
    end
  end

let open_tag c = function
    B ->       {c with b = true}
  | EM ->      if c.s then c else {c with em = not c.em}
  | I ->       {c with i = true}
  | PL ->      {c with s = false; em = false; i = false;
                       b = false; tt = false; u = 0}
  | S ->       {c with s = true; em = false}
  | TT ->      {c with tt = true}
  | U ->       if c.u = 3 then c else {c with u = c.u + 1}
  | Size s ->  {c with size = s}
  | Color l -> {c with color = l}

exception Parse_error of string

let parse_buf lexbuf =
  let rec parse ctx stk =
    match Lexer.token lexbuf with
      Char c ->
        output_char c ctx; parse ctx stk
    | Blank c ->
        output_blank c ctx; parse ctx stk
    | Opening t ->
        let new_ctx = open_tag ctx t in
        parse new_ctx ((t, ctx) :: stk)
    | Closing t ->
        begin match stk with
          (t', ctx') :: stk' when t = t' ->
            parse ctx' stk'
        | _ ->
            raise (Parse_error "bad tag nesting")
        end
    | EOF ->
        if stk <> [] then raise (Parse_error "unclosed tags") else () in
  Buffer.clear text;
  runs := [];
  last_was_space := false;
  last_context := initial_context;
  start_run := 0;
  parse initial_context [];
  add_new_run();
  let t = Buffer.contents text and r = List.rev !runs in
  Buffer.reset text;
  runs := [];
  (t, r)

let parse_channel ic =
  parse_buf (Lexing.from_channel ic)

let parse_file filename =
  let ic = open_in filename in
  let res = parse_channel ic in
  close_in ic;
  res

let parse_string s =
  parse_buf (Lexing.from_string s)
