(* Compare two files for equivalence *)

open Lexer
open Parser

type state = { ctx: context;
               stack: (tag * context) list;
               last_space: context option }

let initial_state = { ctx = initial_context; stack = []; last_space = None }

exception Parse_error

(* Read tokens from lexbuf until a character is output.  Return the
   character and the next state.  End-of-file is represented by
   the special character eof. *)

let eof = '\255'

let rec read lexbuf state =
  match Lexer.token lexbuf with
    Char c ->
      (c, {state with last_space = None})
  | Blank c ->
      let c' = if state.ctx.tt then c else ' ' in
      begin match state.last_space with
        None ->
          (c', {state with last_space = Some state.ctx})
      | Some ctx ->
          if state.ctx.tt then
            (c', {state with last_space = Some state.ctx})
          else if same_space_context ctx state.ctx then
            read lexbuf state
          else
            (c', {state with last_space = Some state.ctx})
      end
  | Opening t ->
      read lexbuf {state with ctx = open_tag state.ctx t;
                              stack = (t, state.ctx) :: state.stack }
  | Closing t ->
      begin match state.stack with
        (t', ctx') :: stk' when t = t' ->
          read lexbuf {state with ctx = ctx'; stack = stk'}
      | _ ->
          raise Parse_error
      end
  | EOF ->
      if state.stack <> [] then raise Parse_error;
      (eof, state)

(* Compare two files *)

let agree (c1,st1) (c2,st2) =
  c1 = c2 &&
  (if c1 = ' ' then same_space_context st1.ctx st2.ctx else st1.ctx = st2.ctx)

let equiv lb1 lb2 =
  let rec compare st1 st2 =
    let ((c1, st1') as res1) = read lb1 st1
    and ((c2, st2') as res2) = read lb2 st2 in
    agree res1 res2 && (c1 = eof || compare st1' st2') in
  compare initial_state initial_state

let equivalent f1 f2 =
  let ic1 = open_in f1 and ic2 = open_in f2 in
  try
    let lb1 = Lexing.from_channel ic1 and lb2 = Lexing.from_channel ic2 in
    let res = equiv lb1 lb2 in
    close_in ic1; close_in ic2; 
    res
  with Parse_error ->
    close_in ic1; close_in ic2; 
    false

let equivalent_strings s1 s2 =
  try
    equiv (Lexing.from_string s1) (Lexing.from_string s2)
  with Parse_error ->
    false
